@@ -2,7 +2,7 @@
 
   encoding.c -
 
-  $Author: nobu $
+  $Author: akr $
   created at: Thu May 24 17:23:27 JST 2007
 
   Copyright (C) 2007 Yukihiro Matsumoto
@@ -18,11 +18,13 @@
 #endif
 
 static ID id_encoding, id_base_encoding;
-static VALUE rb_cEncoding;
+VALUE rb_cEncoding;
+static VALUE rb_encoding_list;
 
 struct rb_encoding_entry {
     const char *name;
     rb_encoding *enc;
+    rb_encoding *base;
 };
 
 static struct {
@@ -34,40 +36,10 @@ static struct {
 
 void rb_enc_init(void);
 
-#ifndef NO_ENCDB_H
-#undef ENC_REPLICATE
-#undef ENC_ALIAS
-#undef ENC_DUMMY
-static int encdb_replicate(const char *alias, const char *orig);
-static int encdb_alias(const char *alias, const char *orig);
-static int encdb_dummy(const char *name);
-static void encdb_declare(const char *name);
-#define ENC_REPLICATE(name, orig) encdb_replicate(name, orig)
-#define ENC_ALIAS(name, orig) encdb_alias(name, orig)
-#define ENC_DUMMY(name) encdb_dummy(name)
-#define ENC_DEFINE(name) encdb_declare(name)
-#endif
-
-static void
-enc_init_db(void)
-{
-#ifdef NO_ENCDB_H
 #define ENCODING_COUNT ENCINDEX_BUILTIN_MAX
-#else
-#include "encdb.h"
-#endif
-}
 
 #define enc_autoload_p(enc) (!rb_enc_mbmaxlen(enc))
 
-#define ENC_UNINITIALIZED (&rb_cEncoding)
-#define enc_initialized_p(enc) ((enc)->auxiliary_data != &rb_cEncoding)
-#define ENC_FROM_ENCODING(enc) ((VALUE)(enc)->auxiliary_data)
-
-#define ENC_DUMMY_FLAG FL_USER2
-#define ENC_DUMMY_P(enc) (RBASIC(enc)->flags & ENC_DUMMY_FLAG)
-#define ENC_SET_DUMMY(enc) (RBASIC(enc)->flags |= ENC_DUMMY_FLAG)
-
 static int load_encoding(const char *name);
 static VALUE enc_base_encoding(VALUE self);
 
@@ -79,36 +51,60 @@ enc_mark(void *ptr)
 static VALUE
 enc_new(rb_encoding *encoding)
 {
-    VALUE enc = Data_Wrap_Struct(rb_cEncoding, enc_mark, 0, encoding);
-    encoding->auxiliary_data = (void *)enc;
-    return enc;
+    return Data_Wrap_Struct(rb_cEncoding, enc_mark, 0, encoding);
 }
 
 VALUE
 rb_enc_from_encoding(rb_encoding *encoding)
 {
+    VALUE list, enc;
+    int idx;
+
     if (!encoding) return Qnil;
-    if (enc_initialized_p(encoding))
-	return ENC_FROM_ENCODING(encoding);
-    return enc_new(encoding);
+    idx = ENC_TO_ENCINDEX(encoding);
+    if (!(list = rb_encoding_list)) {
+	rb_bug("rb_enc_from_encoding(%d\"%s\"): no rb_encoding_list",
+	       idx, rb_enc_name(encoding));
+    }
+    enc = rb_ary_entry(list, idx);
+    if (NIL_P(enc)) {
+	rb_bug("rb_enc_from_encoding(%d\"%s\"): not created yet",
+	       idx, rb_enc_name(encoding));
+    }
+    return enc;
+}
+
+static int enc_autoload(rb_encoding *);
+
+static int
+check_encoding(rb_encoding *enc)
+{
+    int index = rb_enc_to_index(enc);
+    if (rb_enc_from_index(index) != enc)
+	return -1;
+    if (enc_autoload_p(enc)) {
+	index = enc_autoload(enc);
+    }
+    return index;
 }
 
 static int
 enc_check_encoding(VALUE obj)
 {
-    int index;
-    rb_encoding *enc;
-
     if (SPECIAL_CONST_P(obj) || BUILTIN_TYPE(obj) != T_DATA ||
 	RDATA(obj)->dmark != enc_mark) {
 	return -1;
     }
-    enc = (rb_encoding*)RDATA(obj)->data;
-    index = rb_enc_to_index(enc);
-    if (rb_enc_from_index(index) != enc)
-	return -1;
-    if (enc_autoload_p(enc)) {
-	index = load_encoding(enc->name);
+    return check_encoding(RDATA(obj)->data);
+}
+
+static int
+must_encoding(VALUE enc)
+{
+    int index = enc_check_encoding(enc);
+    if (index < 0) {
+	rb_raise(rb_eTypeError, "wrong argument type %s (expected Encoding)",
+		 rb_obj_classname(enc));
     }
     return index;
 }
@@ -146,13 +142,6 @@ rb_to_encoding(VALUE enc)
 void
 rb_gc_mark_encodings(void)
 {
-    int i;
-    for (i = 0; i < enc_table.count; ++i) {
-	rb_encoding *enc = enc_table.list[i].enc;
-	if (enc && enc_initialized_p(enc)) {
-	    rb_gc_mark(ENC_FROM_ENCODING(enc));
-	}
-    }
 }
 
 static int
@@ -175,7 +164,7 @@ static int
 enc_register_at(int index, const char *name, rb_encoding *encoding)
 {
     struct rb_encoding_entry *ent = &enc_table.list[index];
-    void *obj = ENC_UNINITIALIZED;
+    VALUE list;
 
     if (!ent->name) {
 	ent->name = name = strdup(name);
@@ -184,10 +173,7 @@ enc_register_at(int index, const char *name, rb_encoding *encoding)
 	return -1;
     }
     if (!ent->enc) {
-	ent->enc = malloc(sizeof(rb_encoding));
-    }
-    else {
-	obj = ent->enc->auxiliary_data;
+	ent->enc = xmalloc(sizeof(rb_encoding));
     }
     if (encoding) {
 	*ent->enc = *encoding;
@@ -199,15 +185,10 @@ enc_register_at(int index, const char *name, rb_encoding *encoding)
     encoding->name = name;
     encoding->ruby_encoding_index = index;
     st_insert(enc_table.names, (st_data_t)name, (st_data_t)index);
-    if (obj != ENC_UNINITIALIZED) {
-	encoding->auxiliary_data = obj;
-    }
-    else if (rb_cEncoding) {
+    list = rb_encoding_list;
+    if (list && NIL_P(rb_ary_entry(list, index))) {
 	/* initialize encoding data */
-	enc_new(encoding);
-    }
-    else {
-	encoding->auxiliary_data = ENC_UNINITIALIZED;
+	rb_ary_store(list, index, enc_new(encoding));
     }
     return index;
 }
@@ -235,9 +216,7 @@ rb_enc_register(const char *name, rb_encoding *encoding)
 	if (STRCASECMP(name, rb_enc_name(oldenc))) {
 	    index = enc_register(name, encoding);
 	}
-	else if (!enc_autoload_p(oldenc) ||
-		 (enc_initialized_p(oldenc) &&
-		  !ENC_DUMMY_P(ENC_FROM_ENCODING(oldenc)))) {
+	else if (enc_autoload_p(oldenc) || !ENC_DUMMY_P(oldenc)) {
 	    enc_register_at(index, name, encoding);
 	}
 	else {
@@ -251,9 +230,8 @@ rb_enc_register(const char *name, rb_encoding *encoding)
     return index;
 }
 
-#ifndef NO_ENCDB_H
-static void
-encdb_declare(const char *name)
+void
+rb_encdb_declare(const char *name)
 {
     int idx = rb_enc_registered(name);
     if (idx < 0) {
@@ -261,7 +239,6 @@ encdb_declare(const char *name)
     }
     set_encoding_const(name, rb_enc_from_index(idx));
 }
-#endif
 
 static void
 enc_check_duplication(const char *name)
@@ -271,12 +248,12 @@ enc_check_duplication(const char *name)
     }
 }
 
-static VALUE
+static rb_encoding*
 set_base_encoding(int index, rb_encoding *base)
 {
-    VALUE enc = rb_enc_from_encoding(enc_table.list[index].enc);
+    rb_encoding *enc = enc_table.list[index].enc;
 
-    rb_ivar_set(enc, id_base_encoding, rb_enc_from_encoding(base));
+    enc_table.list[index].base = base;
     if (rb_enc_dummy_p(base)) ENC_SET_DUMMY(enc);
     return enc;
 }
@@ -293,7 +270,6 @@ rb_enc_replicate(const char *name, rb_encoding *encoding)
     return idx;
 }
 
-#ifndef NO_ENCDB_H
 static int
 enc_replicate(int idx, const char *name, rb_encoding *origenc)
 {
@@ -310,8 +286,8 @@ enc_replicate(int idx, const char *name, rb_encoding *origenc)
     return idx;
 }
 
-static int
-encdb_replicate(const char *name, const char *orig)
+int
+rb_encdb_replicate(const char *name, const char *orig)
 {
     int origidx = rb_enc_registered(orig);
     int idx = rb_enc_registered(name);
@@ -321,39 +297,27 @@ encdb_replicate(const char *name, const char *orig)
     }
     return enc_replicate(idx, name, rb_enc_from_index(origidx));
 }
-#endif
 
 int
 rb_define_dummy_encoding(const char *name)
 {
     int index = rb_enc_replicate(name, rb_ascii8bit_encoding());
-    VALUE enc = rb_enc_from_encoding(enc_table.list[index].enc);
+    rb_encoding *enc = enc_table.list[index].enc;
 
     ENC_SET_DUMMY(enc);
     return index;
 }
 
-#ifndef NO_ENCDB_H
-static int
-encdb_dummy(const char *name)
+int
+rb_encdb_dummy(const char *name)
 {
     int index = enc_replicate(rb_enc_registered(name), name,
 			      rb_ascii8bit_encoding());
-    VALUE enc = rb_enc_from_encoding(enc_table.list[index].enc);
+    rb_encoding *enc = enc_table.list[index].enc;
 
     ENC_SET_DUMMY(enc);
     return index;
 }
-#endif
-
-int
-rb_enc_dummy_p(rb_encoding *enc)
-{
-    VALUE encoding;
-    if (!enc_initialized_p(enc)) return Qfalse;
-    encoding = rb_enc_from_encoding(enc);
-    return ENC_DUMMY_P(encoding);
-}
 
 /*
  * call-seq:
@@ -371,7 +335,7 @@ rb_enc_dummy_p(rb_encoding *enc)
 static VALUE
 enc_dummy_p(VALUE enc)
 {
-    return rb_enc_dummy_p(rb_to_encoding(enc)) ? Qtrue : Qfalse;
+    return ENC_DUMMY_P(enc_table.list[must_encoding(enc)].enc) ? Qtrue : Qfalse;
 }
 
 static int
@@ -398,9 +362,8 @@ rb_enc_alias(const char *alias, const char *orig)
     return enc_alias(alias, idx);
 }
 
-#ifndef NO_ENCDB_H
-static int
-encdb_alias(const char *alias, const char *orig)
+int
+rb_encdb_alias(const char *alias, const char *orig)
 {
     int idx = rb_enc_registered(orig);
 
@@ -409,7 +372,6 @@ encdb_alias(const char *alias, const char *orig)
     }
     return enc_alias(alias, idx);
 }
-#endif
 
 enum {
     ENCINDEX_ASCII,
@@ -495,32 +457,43 @@ load_encoding(const char *name)
     return idx;
 }
 
+static int
+enc_autoload(rb_encoding *enc)
+{
+    int i;
+    rb_encoding *base = enc_table.list[ENC_TO_ENCINDEX(enc)].base;
+
+    if (base) {
+	i = 0;
+	do {
+	    if (i >= enc_table.count) return -1;
+	} while (enc_table.list[i].enc != base && (++i, 1));
+	if (enc_autoload_p(base)) {
+	    if (enc_autoload(base) < 0) return -1;
+	}
+	i = ENC_TO_ENCINDEX(enc);
+	enc_register_at(i, rb_enc_name(enc), base);
+    }
+    else {
+	i = load_encoding(rb_enc_name(enc));
+    }
+    return i;
+}
+
 int
 rb_enc_find_index(const char *name)
 {
-    int i = rb_enc_registered(name), b;
+    int i = rb_enc_registered(name);
     rb_encoding *enc;
-    VALUE base;
 
     if (i < 0) {
 	i = load_encoding(name);
     }
     else if (enc_autoload_p(enc = rb_enc_from_index(i))) {
-	if (enc_initialized_p(enc) &&
-	    (base = enc_base_encoding(ENC_FROM_ENCODING(enc)), !NIL_P(base))) {
-	    if ((b = enc_check_encoding(base)) < 0) {
-		goto failed;
-	    }
-	    enc_register_at(i, rb_enc_name(enc), rb_enc_from_index(b));
-	}
-	else {
-	    i = load_encoding(rb_enc_name(enc));
-	    if (i < 0) {
-	      failed:
-		rb_warn("failed to load encoding (%s); use ASCII-8BIT instead",
-			name);
-		return 0;
-	    }
+	if (enc_autoload(enc) < 0) {
+	    rb_warn("failed to load encoding (%s); use ASCII-8BIT instead",
+		    name);
+	    return 0;
 	}
     }
     return i;
@@ -550,42 +523,15 @@ enc_capable(VALUE obj)
     }
 }
 
-static void
-enc_check_capable(VALUE x)
-{
-    if (!enc_capable(x)) {
-	const char *etype;
-
-	if (NIL_P(x)) {
-	    etype = "nil";
-	}
-	else if (FIXNUM_P(x)) {
-	    etype = "Fixnum";
-	}
-	else if (SYMBOL_P(x)) {
-	    etype = "Symbol";
-	}
-	else if (rb_special_const_p(x)) {
-	    etype = RSTRING_PTR(rb_obj_as_string(x));
-	}
-	else {
-	    etype = rb_obj_classname(x);
-	}
-	rb_raise(rb_eTypeError, "wrong argument type %s (not encode capable)", etype);
-    }
-}
-
 ID
 rb_id_encoding(void)
 {
-    if (!id_encoding) {
-	id_encoding = rb_intern("encoding");
-    }
+    CONST_ID(id_encoding, "encoding");
     return id_encoding;
 }
 
 int
-rb_enc_internal_get_index(VALUE obj)
+rb_enc_get_index(VALUE obj)
 {
     int i;
 
@@ -600,7 +546,7 @@ rb_enc_internal_get_index(VALUE obj)
 }
 
 void
-rb_enc_internal_set_index(VALUE obj, int idx)
+rb_enc_set_index(VALUE obj, int idx)
 {
     if (idx < ENCODING_INLINE_MAX) {
 	ENCODING_SET_INLINED(obj, idx);
@@ -611,30 +557,24 @@ rb_enc_internal_set_index(VALUE obj, int idx)
     return;
 }
 
-void
+VALUE
 rb_enc_associate_index(VALUE obj, int idx)
 {
-    enc_check_capable(obj);
+/*    enc_check_capable(obj);*/
     if (rb_enc_get_index(obj) == idx)
-    	return;
+    	return obj;
     if (!ENC_CODERANGE_ASCIIONLY(obj) ||
 	!rb_enc_asciicompat(rb_enc_from_index(idx))) {
 	ENC_CODERANGE_CLEAR(obj);
     }
-    rb_enc_internal_set_index(obj, idx);
+    rb_enc_set_index(obj, idx);
+    return obj;
 }
 
-void
+VALUE
 rb_enc_associate(VALUE obj, rb_encoding *enc)
 {
-    rb_enc_associate_index(obj, rb_enc_to_index(enc));
-}
-
-int
-rb_enc_get_index(VALUE obj)
-{
-    if (!enc_capable(obj)) return -1;
-    return rb_enc_internal_get_index(obj);
+    return rb_enc_associate_index(obj, rb_enc_to_index(enc));
 }
 
 rb_encoding*
@@ -830,7 +770,7 @@ enc_inspect(VALUE self)
 {
     VALUE str = rb_sprintf("#<%s:%s%s>", rb_obj_classname(self),
 		      rb_enc_name((rb_encoding*)DATA_PTR(self)),
-		      (ENC_DUMMY_P(self) ? " (dummy)" : ""));
+		      (enc_dummy_p(self) ? " (dummy)" : ""));
     ENCODING_CODERANGE_SET(str, rb_usascii_encindex(), ENC_CODERANGE_7BIT);
     return str;
 }
@@ -852,7 +792,9 @@ enc_name(VALUE self)
 static VALUE
 enc_base_encoding(VALUE self)
 {
-    return rb_attr_get(self, id_base_encoding);
+    rb_encoding *base = enc_table.list[must_encoding(self)].base;
+    if (!base) return Qnil;
+    return ENC_FROM_ENCODING(base);
 }
 
 /*
@@ -876,14 +818,8 @@ enc_base_encoding(VALUE self)
 static VALUE
 enc_list(VALUE klass)
 {
-    VALUE ary = rb_ary_new2(enc_table.count);
-    int i;
-    for (i = 0; i < enc_table.count; ++i) {
-	rb_encoding *enc = enc_table.list[i].enc;
-	if (enc) {
-	    rb_ary_push(ary, rb_enc_from_encoding(enc));
-	}
-    }
+    VALUE ary = rb_ary_new2(0);
+    rb_ary_replace(ary, rb_encoding_list);
     return ary;
 }
 
@@ -936,11 +872,13 @@ enc_find(VALUE klass, VALUE enc)
 static VALUE
 enc_compatible_p(VALUE klass, VALUE str1, VALUE str2)
 {
-    rb_encoding *enc = rb_enc_compatible(str1, str2);
-    VALUE encoding = Qnil;
-    if (!enc || !(encoding = rb_enc_from_encoding(enc)))
-	encoding = Qnil;
-    return encoding;
+    rb_encoding *enc;
+
+    if (!enc_capable(str1)) return Qnil;
+    if (!enc_capable(str2)) return Qnil;
+    enc = rb_enc_compatible(str1, str2);
+    if (!enc) return Qnil;
+    return rb_enc_from_encoding(enc);
 }
 
 /* :nodoc: */
@@ -964,7 +902,13 @@ rb_ascii8bit_encoding(void)
     if (!enc_table.list) {
 	rb_enc_init();
     }
-    return enc_table.list[0].enc;
+    return enc_table.list[ENCINDEX_ASCII].enc;
+}
+
+int
+rb_ascii8bit_encindex(void)
+{
+    return ENCINDEX_ASCII;
 }
 
 rb_encoding *
@@ -976,6 +920,12 @@ rb_utf8_encoding(void)
     return enc_table.list[ENCINDEX_UTF_8].enc;
 }
 
+int
+rb_utf8_encindex(void)
+{
+    return ENCINDEX_UTF_8;
+}
+
 rb_encoding *
 rb_usascii_encoding(void)
 {
@@ -998,7 +948,7 @@ rb_locale_encoding(void)
     int idx;
 
     if (NIL_P(charmap))
-        idx = rb_enc_find_index("US-ASCII");
+        return rb_usascii_encoding();
     else
         idx = rb_enc_find_index(StringValueCStr(charmap));
     if (idx < 0)
@@ -1007,6 +957,22 @@ rb_locale_encoding(void)
     return rb_enc_from_index(idx);
 }
 
+rb_encoding *
+rb_filesystem_encoding(void)
+{
+    static rb_encoding *enc;
+    if (!enc) {
+#if defined _WIN32
+	enc = rb_locale_encoding();
+#elif defined __APPLE__
+	enc = rb_enc_find("UTF8-MAC");
+#else
+	enc = rb_locale_encoding();
+#endif
+    }
+    return enc;
+}
+
 static int default_external_index;
 
 rb_encoding *
@@ -1205,6 +1171,10 @@ rb_enc_aliases(VALUE klass)
 void
 Init_Encoding(void)
 {
+#undef rb_intern
+    VALUE list;
+    int i;
+
     id_base_encoding = rb_intern("#base_encoding");
 
     rb_cEncoding = rb_define_class("Encoding", rb_cObject);
@@ -1226,7 +1196,13 @@ Init_Encoding(void)
     rb_define_singleton_method(rb_cEncoding, "default_external", get_default_external, 0);
     rb_define_singleton_method(rb_cEncoding, "locale_charmap", rb_locale_charmap, 0);
 
-    enc_init_db();
+    rb_gc_register_address(&rb_encoding_list);
+    list = rb_ary_new2(enc_table.count);
+    RBASIC(list)->klass = 0;
+    rb_encoding_list = list;
+    for (i = 0; i < enc_table.count; ++i) {
+	rb_ary_push(list, enc_new(enc_table.list[i].enc));
+    }
 }
 
 /* locale insensitive functions */
