@@ -4,11 +4,11 @@
  * <amos+ruby@utdallas.edu>
  *
  * $RoughId: syslog.c,v 1.21 2002/02/25 12:21:17 knu Exp $
- * $Id: syslog.c 35746 2012-05-21 07:35:21Z knu $
+ * $Id: syslog.c 17726 2008-06-30 01:30:32Z matz $
  */
 
-#include "ruby.h"
-#include "util.h"
+#include "ruby/ruby.h"
+#include "ruby/util.h"
 #include <syslog.h>
 
 /* Syslog class */
@@ -33,7 +33,7 @@ static void syslog_write(int pri, int argc, VALUE *argv)
 
     str = rb_f_sprintf(argc, argv);
 
-    syslog(pri, "%s", RSTRING(str)->ptr);
+    syslog(pri, "%s", RSTRING_PTR(str));
 }
 
 /* Syslog module methods */
@@ -72,7 +72,7 @@ static VALUE mSyslog_open(int argc, VALUE *argv, VALUE self)
 #else
     Check_SafeStr(ident);
 #endif
-    syslog_ident = strdup(RSTRING(ident)->ptr);
+    syslog_ident = strdup(RSTRING_PTR(ident));
 
     if (NIL_P(opt)) {
 	syslog_options = LOG_PID | LOG_CONS;
@@ -168,8 +168,6 @@ static VALUE mSyslog_inspect(VALUE self)
 {
     char buf[1024];
 
-    Check_Type(self, T_MODULE);
-
     if (syslog_opened) {
 	snprintf(buf, sizeof(buf),
 	  "<#%s: opened=true, ident=\"%s\", options=%d, facility=%d, mask=%d>",
