@@ -2,7 +2,7 @@
 
   id.c - 
 
-  $Author: akr $
+  $Author: ko1 $
   created at: Thu Jul 12 04:37:51 2007
 
   Copyright (C) 2004-2007 Koichi Sasada
@@ -18,6 +18,8 @@
 void
 Init_id(void)
 {
+#undef rb_intern
+
     /* Symbols */
     symIFUNC = ID2SYM(rb_intern("<IFUNC>"));
     symCFUNC = ID2SYM(rb_intern("<CFUNC>"));
@@ -63,4 +65,14 @@ Init_id(void)
 
     idSend = rb_intern("send");
     id__send__ = rb_intern("__send__");
+
+    idRespond_to = rb_intern("respond_to?");
+    idInitialize = rb_intern("initialize");
+
+    id_core_set_method_alias = rb_intern("core_set_method_alias");
+    id_core_set_variable_alias = rb_intern("core_set_variable_alias");
+    id_core_undef_method = rb_intern("core_undef_method");
+    id_core_define_method = rb_intern("core_define_method");
+    id_core_define_singleton_method = rb_intern("core_define_singleton_method");
+    id_core_set_postexe = rb_intern("core_set_postexe");
 }
