@@ -50,5 +50,14 @@ extern ID idBitblt;
 extern ID idAnswer;
 extern ID idSend;
 extern ID id__send__;
+extern ID idRespond_to;
+extern ID idInitialize;
+
+extern ID id_core_set_method_alias;
+extern ID id_core_set_variable_alias;
+extern ID id_core_undef_method;
+extern ID id_core_define_method;
+extern ID id_core_define_singleton_method;
+extern ID id_core_set_postexe;
 
 #endif /* RUBY_ID_H */
