@@ -2,7 +2,7 @@
 
   numeric.c -
 
-  $Author: matz $
+  $Author: usa $
   created at: Fri Aug 13 18:33:09 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -118,7 +118,9 @@ num_coerce(VALUE x, VALUE y)
 {
     if (CLASS_OF(x) == CLASS_OF(y))
 	return rb_assoc_new(y, x);
-    return rb_assoc_new(rb_Float(y), rb_Float(x));
+    x = rb_Float(x);
+    y = rb_Float(y);
+    return rb_assoc_new(y, x);
 }
 
 static VALUE
@@ -249,15 +251,28 @@ num_uminus(VALUE num)
 /*
  *  call-seq:
  *     num.quo(numeric)    =>   result
- *     num.fdiv(numeric)   =>   result
  *
- *  Equivalent to <code>Numeric#/</code>, but overridden in subclasses.
+ *  Returns most exact division (rational for integers, float for floats).
  */
 
 static VALUE
 num_quo(VALUE x, VALUE y)
 {
-    return rb_funcall(x, '/', 1, y);
+    return rb_funcall(rb_rational_raw1(x), '/', 1, y);
+}
+
+
+/*
+ *  call-seq:
+ *     num.fdiv(numeric)    =>   float
+ *
+ *  Returns float division.
+ */
+
+static VALUE
+num_fdiv(VALUE x, VALUE y)
+{
+    return rb_funcall(rb_Float(x), '/', 1, y);
 }
 
 
@@ -275,6 +290,7 @@ static VALUE num_floor(VALUE num);
 static VALUE
 num_div(VALUE x, VALUE y)
 {
+    if (rb_equal(INT2FIX(0), y)) rb_num_zerodiv();
     return num_floor(rb_funcall(x, '/', 1, y));
 }
 
@@ -312,6 +328,7 @@ num_div(VALUE x, VALUE y)
  *
  *
  *  Examples
+ *
  *     11.divmod(3)         #=> [3, 2]
  *     11.divmod(-3)        #=> [-4, -1]
  *     11.divmod(3.5)       #=> [3, 0.5]
@@ -645,6 +662,11 @@ flo_div(VALUE x, VALUE y)
     }
 }
 
+static VALUE
+flo_quo(VALUE x, VALUE y)
+{
+    return rb_funcall(x, '/', 1, y);
+}
 
 static void
 flodivmod(double x, double y, double *divp, double *modp)
@@ -707,6 +729,22 @@ flo_mod(VALUE x, VALUE y)
     return DOUBLE2NUM(mod);
 }
 
+static VALUE
+dbl2ival(double d)
+{
+    if (FIXABLE(d)) {
+	d = round(d);
+	return LONG2FIX((long)d);
+    }
+    else if (isnan(d) || isinf(d)) {
+	/* special case: cannot return integer value */
+	return rb_float_new(d);
+    }
+    else {
+	return rb_dbl2big(d);
+    }
+}
+
 /*
  *  call-seq:
  *     flt.divmod(numeric)    => array
@@ -717,7 +755,7 @@ flo_mod(VALUE x, VALUE y)
 static VALUE
 flo_divmod(VALUE x, VALUE y)
 {
-    double fy, div, mod, val;
+    double fy, div, mod;
     volatile VALUE a, b;
 
     switch (TYPE(y)) {
@@ -734,16 +772,7 @@ flo_divmod(VALUE x, VALUE y)
 	return rb_num_coerce_bin(x, y, rb_intern("divmod"));
     }
     flodivmod(RFLOAT_VALUE(x), fy, &div, &mod);
-    if (FIXABLE(div)) {
-	val = round(div);
-	a = LONG2FIX(val);
-    }
-    else if (isnan(div) || isinf(div)) {
-	a = rb_float_new(div);
-    }
-    else {
-	a = rb_dbl2big(div);
-    }
+    a = dbl2ival(div);
     b = DOUBLE2NUM(mod);
     return rb_assoc_new(a, b);
 }
@@ -1263,7 +1292,7 @@ flo_round(int argc, VALUE *argv, VALUE num)
     int ndigits = 0, i;
     long val;
 
-    if (rb_scan_args(argc, argv, "01", &nd) == 1) {
+    if (argc > 0 && rb_scan_args(argc, argv, "01", &nd) == 1) {
 	ndigits = NUM2INT(nd);
     }
     number  = RFLOAT_VALUE(num);
@@ -1272,11 +1301,16 @@ flo_round(int argc, VALUE *argv, VALUE num)
     while  (--i >= 0)
 	f = f*10.0;
 
-    if (ndigits < 0) number /= f;
-    else number *= f;
-    number = round(number);
-    if (ndigits < 0) number *= f;
-    else number /= f;
+    if (isinf(f)) {
+	if (ndigits < 0) number = 0;
+    }
+    else {
+	if (ndigits < 0) number /= f;
+	else number *= f;
+	number = round(number);
+	if (ndigits < 0) number *= f;
+	else number /= f;
+    }
 
     if (ndigits > 0) return DOUBLE2NUM(number);
 
@@ -1492,6 +1526,7 @@ num_step(int argc, VALUE *argv, VALUE from)
 SIGNED_VALUE
 rb_num2long(VALUE val)
 {
+  again:
     if (NIL_P(val)) {
 	rb_raise(rb_eTypeError, "no implicit conversion from nil to integer");
     }
@@ -1518,7 +1553,7 @@ rb_num2long(VALUE val)
 
       default:
 	val = rb_to_int(val);
-	return NUM2LONG(val);
+	goto again;
     }
 }
 
@@ -1546,22 +1581,23 @@ check_int(SIGNED_VALUE num)
     else {
 	return;
     }
-#ifdef LONG_LONG_VALUE
-    rb_raise(rb_eRangeError, "integer %lld too %s to convert to `int'", num, s);
-#else
-    rb_raise(rb_eRangeError, "integer %ld too %s to convert to `int'", num, s);
-#endif
+    rb_raise(rb_eRangeError, "integer %"PRIdVALUE " too %s to convert to `int'", num, s);
 }
 
 static void
-check_uint(VALUE num)
+check_uint(VALUE num, VALUE sign)
 {
-    if (num > UINT_MAX) {
-#ifdef LONG_LONG_VALUE
-	rb_raise(rb_eRangeError, "integer %llu too big to convert to `unsigned int'", num);
-#else
-	rb_raise(rb_eRangeError, "integer %lu too big to convert to `unsigned int'", num);
-#endif
+    static const VALUE mask = ~(VALUE)UINT_MAX;
+
+    if (RTEST(sign)) {
+	/* minus */
+	if ((num & mask) != mask || (num & ~mask) <= INT_MAX + 1UL)
+	    rb_raise(rb_eRangeError, "integer %"PRIdVALUE " too small to convert to `unsigned int'", num);
+    }
+    else {
+	/* plus */
+	if ((num & mask) != 0)
+	    rb_raise(rb_eRangeError, "integer %"PRIuVALUE " too big to convert to `unsigned int'", num);
     }
 }
 
@@ -1588,9 +1624,7 @@ rb_num2uint(VALUE val)
 {
     unsigned long num = rb_num2ulong(val);
 
-    if (RTEST(rb_funcall(INT2FIX(0), '<', 1, val))) {
-	check_uint(num);
-    }
+    check_uint(num, rb_funcall(val, '<', 1, INT2FIX(0)));
     return num;
 }
 
@@ -1603,9 +1637,8 @@ rb_fix2uint(VALUE val)
 	return rb_num2uint(val);
     }
     num = FIX2ULONG(val);
-    if (FIX2LONG(val) > 0) {
-	check_uint(num);
-    }
+
+    check_uint(num, rb_funcall(val, '<', 1, INT2FIX(0)));
     return num;
 }
 #else
@@ -1631,7 +1664,7 @@ rb_num2fix(VALUE val)
 
     v = rb_num2long(val);
     if (!FIXABLE(v))
-	rb_raise(rb_eRangeError, "integer %ld out of range of fixnum", v);
+	rb_raise(rb_eRangeError, "integer %"PRIdVALUE " out of range of fixnum", v);
     return LONG2FIX(v);
 }
 
@@ -1690,6 +1723,17 @@ rb_num2ull(VALUE val)
 
 #endif  /* HAVE_LONG_LONG */
 
+static VALUE
+num_numerator(VALUE num)
+{
+    return rb_funcall(rb_Rational1(num), rb_intern("numerator"), 0);
+}
+
+static VALUE
+num_denominator(VALUE num)
+{
+    return rb_funcall(rb_Rational1(num), rb_intern("denominator"), 0);
+}
 
 /*
  * Document-class: Integer
@@ -1848,7 +1892,7 @@ int_chr(int argc, VALUE *argv, VALUE num)
       case 0:
 	if (i < 0 || 0xff < i) {
 	  out_of_range:
-	    rb_raise(rb_eRangeError, "%ld out of char range", i);
+	    rb_raise(rb_eRangeError, "%"PRIdVALUE " out of char range", i);
 	}
 	c = i;
 	if (i < 0x80) {
@@ -1871,6 +1915,18 @@ int_chr(int argc, VALUE *argv, VALUE num)
     return str;
 }
 
+static VALUE
+int_numerator(VALUE num)
+{
+    return num;
+}
+
+static VALUE
+int_denominator(VALUE num)
+{
+    return INT2FIX(1);
+}
+
 /********************************************************************
  *
  * Document-class: Fixnum
@@ -1919,6 +1975,7 @@ rb_int_induced_from(VALUE klass, VALUE x)
       case T_BIGNUM:
 	return x;
       case T_FLOAT:
+      case T_RATIONAL:
 	return rb_funcall(x, id_to_i, 0);
       default:
 	rb_raise(rb_eTypeError, "failed to convert %s into Integer",
@@ -1939,6 +1996,7 @@ rb_flo_induced_from(VALUE klass, VALUE x)
     switch (TYPE(x)) {
       case T_FIXNUM:
       case T_BIGNUM:
+      case T_RATIONAL:
 	return rb_funcall(x, rb_intern("to_f"), 0);
       case T_FLOAT:
 	return x;
@@ -2008,12 +2066,15 @@ rb_fix2str(VALUE x, int base)
 static VALUE
 fix_to_s(int argc, VALUE *argv, VALUE x)
 {
-    VALUE b;
     int base;
 
-    rb_scan_args(argc, argv, "01", &b);
     if (argc == 0) base = 10;
-    else base = NUM2INT(b);
+    else {
+	VALUE b;
+
+	rb_scan_args(argc, argv, "01", &b);
+	base = NUM2INT(b);
+    }
 
     return rb_fix2str(x, base);
 }
@@ -2173,19 +2234,18 @@ fixdivmod(long x, long y, long *divp, long *modp)
 
 /*
  *  call-seq:
- *     fix.quo(numeric)    => float
  *     fix.fdiv(numeric)   => float
  *
  *  Returns the floating point result of dividing <i>fix</i> by
  *  <i>numeric</i>.
  *
- *     654321.quo(13731)      #=> 47.6528293642124
- *     654321.quo(13731.24)   #=> 47.6519964693647
+ *     654321.fdiv(13731)      #=> 47.6528293642124
+ *     654321.fdiv(13731.24)   #=> 47.6519964693647
  *
  */
 
 static VALUE
-fix_quo(VALUE x, VALUE y)
+fix_fdiv(VALUE x, VALUE y)
 {
     if (FIXNUM_P(y)) {
 	return DOUBLE2NUM((double)FIX2LONG(x) / (double)FIX2LONG(y));
@@ -2196,12 +2256,12 @@ fix_quo(VALUE x, VALUE y)
       case T_FLOAT:
 	return DOUBLE2NUM((double)FIX2LONG(x) / RFLOAT_VALUE(y));
       default:
-	return rb_num_coerce_bin(x, y, rb_intern("quo"));
+	return rb_num_coerce_bin(x, y, rb_intern("fdiv"));
     }
 }
 
 static VALUE
-fix_divide(VALUE x, VALUE y, int flo)
+fix_divide(VALUE x, VALUE y, ID op)
 {
     if (FIXNUM_P(y)) {
 	long div;
@@ -2214,15 +2274,21 @@ fix_divide(VALUE x, VALUE y, int flo)
 	x = rb_int2big(FIX2LONG(x));
 	return rb_big_div(x, y);
       case T_FLOAT:
-	if (flo) {
-	    return DOUBLE2NUM((double)FIX2LONG(x) / RFLOAT_VALUE(y));
-	}
-	else {
-	    long div = (double)FIX2LONG(x) / RFLOAT_VALUE(y);
-	    return LONG2NUM(div);
+	{
+	    double div;
+
+	    if (op == '/') {
+		div = (double)FIX2LONG(x) / RFLOAT_VALUE(y);
+		return DOUBLE2NUM(div);
+	    }
+	    else {
+		if (RFLOAT_VALUE(y) == 0) rb_num_zerodiv();
+		div = (double)FIX2LONG(x) / RFLOAT_VALUE(y);
+		return rb_dbl2big(floor(div));
+	    }
 	}
       default:
-	return rb_num_coerce_bin(x, y, flo ? '/' : rb_intern("div"));
+	return rb_num_coerce_bin(x, y, op);
     }
 }
 
@@ -2238,7 +2304,7 @@ fix_divide(VALUE x, VALUE y, int flo)
 static VALUE
 fix_div(VALUE x, VALUE y)
 {
-    return fix_divide(x, y, Qtrue);
+    return fix_divide(x, y, '/');
 }
 
 /*
@@ -2251,7 +2317,7 @@ fix_div(VALUE x, VALUE y)
 static VALUE
 fix_idiv(VALUE x, VALUE y)
 {
-    return fix_divide(x, y, Qfalse);
+    return fix_divide(x, y, rb_intern("div"));
 }
 
 /*
@@ -2314,7 +2380,7 @@ fix_divmod(VALUE x, VALUE y)
 	    volatile VALUE a, b;
 
 	    flodivmod((double)FIX2LONG(x), RFLOAT_VALUE(y), &div, &mod);
-	    a = DOUBLE2NUM(div);
+	    a = dbl2ival(div);
 	    b = DOUBLE2NUM(mod);
 	    return rb_assoc_new(a, b);
 	}
@@ -2380,6 +2446,9 @@ fix_pow(VALUE x, VALUE y)
     if (FIXNUM_P(y)) {
 	long b = FIX2LONG(y);
 
+	if (b < 0)
+	  return rb_funcall(rb_rational_raw1(x), rb_intern("**"), 1, y);
+
 	if (b == 0) return INT2FIX(1);
 	if (b == 1) return x;
 	if (a == 0) {
@@ -2393,13 +2462,14 @@ fix_pow(VALUE x, VALUE y)
 	    else 
 		return INT2FIX(-1);
 	}
-	if (b > 0) {
-	    return int_pow(a, b);
-	}
-	return DOUBLE2NUM(pow((double)a, (double)b));
+	return int_pow(a, b);
     }
     switch (TYPE(y)) {
       case T_BIGNUM:
+
+	if (rb_funcall(y, '<', 1, INT2FIX(0)))
+	  return rb_funcall(rb_rational_raw1(x), rb_intern("**"), 1, y);
+
 	if (a == 0) return INT2FIX(0);
 	if (a == 1) return INT2FIX(1);
 	if (a == -1) {
@@ -2589,6 +2659,18 @@ fix_rev(VALUE num)
     return LONG2NUM(val);
 }
 
+static VALUE
+bit_coerce(VALUE x)
+{
+    while (!FIXNUM_P(x) && TYPE(x) != T_BIGNUM) {
+	if (TYPE(x) == T_FLOAT) {
+	    rb_raise(rb_eTypeError, "can't convert Float into Integer");
+	}
+	x = rb_to_int(x);
+    }
+    return x;
+}
+
 /*
  * call-seq:
  *   fix & other     => integer
@@ -2601,10 +2683,10 @@ fix_and(VALUE x, VALUE y)
 {
     long val;
 
-    if (TYPE(y) == T_BIGNUM) {
+    if (!FIXNUM_P(y = bit_coerce(y))) {
 	return rb_big_and(y, x);
     }
-    val = FIX2LONG(x) & NUM2LONG(y);
+    val = FIX2LONG(x) & FIX2LONG(y);
     return LONG2NUM(val);
 }
 
@@ -2620,10 +2702,10 @@ fix_or(VALUE x, VALUE y)
 {
     long val;
 
-    if (TYPE(y) == T_BIGNUM) {
+    if (!FIXNUM_P(y = bit_coerce(y))) {
 	return rb_big_or(y, x);
     }
-    val = FIX2LONG(x) | NUM2LONG(y);
+    val = FIX2LONG(x) | FIX2LONG(y);
     return LONG2NUM(val);
 }
 
@@ -2639,10 +2721,10 @@ fix_xor(VALUE x, VALUE y)
 {
     long val;
 
-    if (TYPE(y) == T_BIGNUM) {
+    if (!FIXNUM_P(y = bit_coerce(y))) {
 	return rb_big_xor(y, x);
     }
-    val = FIX2LONG(x) ^ NUM2LONG(y);
+    val = FIX2LONG(x) ^ FIX2LONG(y);
     return LONG2NUM(val);
 }
 
@@ -2736,7 +2818,8 @@ fix_aref(VALUE fix, VALUE idx)
     long val = FIX2LONG(fix);
     long i;
 
-    if (TYPE(idx) == T_BIGNUM) {
+    idx = rb_to_int(idx);
+    if (!FIXNUM_P(idx)) {
 	idx = rb_big_norm(idx);
 	if (!FIXNUM_P(idx)) {
 	    if (!RBIGNUM_SIGN(idx) || val >= 0)
@@ -2744,7 +2827,7 @@ fix_aref(VALUE fix, VALUE idx)
 	    return INT2FIX(1);
 	}
     }
-    i = NUM2LONG(idx);
+    i = FIX2LONG(idx);
 
     if (i < 0) return INT2FIX(0);
     if (SIZEOF_LONG*CHAR_BIT-1 < i) {
@@ -2795,52 +2878,6 @@ fix_abs(VALUE fix)
     return LONG2NUM(i);
 }
 
-/*
- *  call-seq:
- *     fix.id2name -> string or nil
- *
- *  Returns the name of the object whose symbol id is <i>fix</i>. If
- *  there is no symbol in the symbol table with this value, returns
- *  <code>nil</code>. <code>id2name</code> has nothing to do with the
- *  <code>Object.id</code> method. See also <code>Fixnum#to_sym</code>,
- *  <code>String#intern</code>, and class <code>Symbol</code>.
- *
- *     symbol = :@inst_var    #=> :@inst_var
- *     id     = symbol.to_i   #=> 9818
- *     id.id2name             #=> "@inst_var"
- */
-
-static VALUE
-fix_id2name(VALUE fix)
-{
-    VALUE name = rb_id2str(FIX2UINT(fix));
-    if (name) return rb_str_dup(name);
-    return Qnil;
-}
-
-
-/*
- *  call-seq:
- *     fix.to_sym -> aSymbol
- *
- *  Returns the symbol whose integer value is <i>fix</i>. See also
- *  <code>Fixnum#id2name</code>.
- *
- *     fred = :fred.to_i
- *     fred.id2name   #=> "fred"
- *     fred.to_sym    #=> :fred
- */
-
-static VALUE
-fix_to_sym(VALUE fix)
-{
-    ID id = FIX2UINT(fix);
-
-    if (rb_id2name(id)) {
-	return ID2SYM(id);
-    }
-    return Qnil;
-}
 
 
 /*
@@ -3068,6 +3105,8 @@ fix_even_p(VALUE num)
 void
 Init_Numeric(void)
 {
+#undef rb_intern
+
 #if defined(__FreeBSD__) && __FreeBSD__ < 4
     /* allow divide by zero -- Inf */
     fpsetmask(fpgetmask() & ~(FP_X_DZ|FP_X_INV|FP_X_OFL));
@@ -3096,7 +3135,7 @@ Init_Numeric(void)
     rb_define_method(rb_cNumeric, "<=>", num_cmp, 1);
     rb_define_method(rb_cNumeric, "eql?", num_eql, 1);
     rb_define_method(rb_cNumeric, "quo", num_quo, 1);
-    rb_define_method(rb_cNumeric, "fdiv", num_quo, 1);
+    rb_define_method(rb_cNumeric, "fdiv", num_fdiv, 1);
     rb_define_method(rb_cNumeric, "div", num_div, 1);
     rb_define_method(rb_cNumeric, "divmod", num_divmod, 1);
     rb_define_method(rb_cNumeric, "modulo", num_modulo, 1);
@@ -3115,6 +3154,9 @@ Init_Numeric(void)
     rb_define_method(rb_cNumeric, "truncate", num_truncate, 0);
     rb_define_method(rb_cNumeric, "step", num_step, -1);
 
+    rb_define_method(rb_cNumeric, "numerator", num_numerator, 0);
+    rb_define_method(rb_cNumeric, "denominator", num_denominator, 0);
+
     rb_cInteger = rb_define_class("Integer", rb_cNumeric);
     rb_undef_alloc_func(rb_cInteger);
     rb_undef_method(CLASS_OF(rb_cInteger), "new");
@@ -3142,10 +3184,10 @@ Init_Numeric(void)
     rb_define_singleton_method(rb_cFixnum, "induced_from", rb_fix_induced_from, 1);
     rb_define_singleton_method(rb_cInteger, "induced_from", rb_int_induced_from, 1);
 
-    rb_define_method(rb_cFixnum, "to_s", fix_to_s, -1);
+    rb_define_method(rb_cInteger, "numerator", int_numerator, 0);
+    rb_define_method(rb_cInteger, "denominator", int_denominator, 0);
 
-    rb_define_method(rb_cFixnum, "id2name", fix_id2name, 0);
-    rb_define_method(rb_cFixnum, "to_sym", fix_to_sym, 0);
+    rb_define_method(rb_cFixnum, "to_s", fix_to_s, -1);
 
     rb_define_method(rb_cFixnum, "-@", fix_uminus, 0);
     rb_define_method(rb_cFixnum, "+", fix_plus, 1);
@@ -3156,8 +3198,7 @@ Init_Numeric(void)
     rb_define_method(rb_cFixnum, "%", fix_mod, 1);
     rb_define_method(rb_cFixnum, "modulo", fix_mod, 1);
     rb_define_method(rb_cFixnum, "divmod", fix_divmod, 1);
-    rb_define_method(rb_cFixnum, "quo", fix_quo, 1);
-    rb_define_method(rb_cFixnum, "fdiv", fix_quo, 1);
+    rb_define_method(rb_cFixnum, "fdiv", fix_fdiv, 1);
     rb_define_method(rb_cFixnum, "**", fix_pow, 1);
 
     rb_define_method(rb_cFixnum, "abs", fix_abs, 0);
@@ -3212,6 +3253,8 @@ Init_Numeric(void)
     rb_define_method(rb_cFloat, "-", flo_minus, 1);
     rb_define_method(rb_cFloat, "*", flo_mul, 1);
     rb_define_method(rb_cFloat, "/", flo_div, 1);
+    rb_define_method(rb_cFloat, "quo", flo_quo, 1);
+    rb_define_method(rb_cFloat, "fdiv", flo_quo, 1);
     rb_define_method(rb_cFloat, "%", flo_mod, 1);
     rb_define_method(rb_cFloat, "modulo", flo_mod, 1);
     rb_define_method(rb_cFloat, "divmod", flo_divmod, 1);
