@@ -2,18 +2,19 @@
 
   signal.c -
 
-  $Author: shyouhei $
-  $Date: 2008-06-15 19:05:11 +0900 (Sun, 15 Jun 2008) $
+  $Author: nobu $
   created at: Tue Dec 20 10:13:44 JST 1994
 
-  Copyright (C) 1993-2003 Yukihiro Matsumoto
+  Copyright (C) 1993-2007 Yukihiro Matsumoto
   Copyright (C) 2000  Network Applied Communication Laboratory, Inc.
   Copyright (C) 2000  Information-technology Promotion Agency, Japan
 
 **********************************************************************/
 
-#include "ruby.h"
-#include "rubysig.h"
+#include "ruby/ruby.h"
+#include "ruby/signal.h"
+#include "ruby/node.h"
+#include "vm_core.h"
 #include <signal.h>
 #include <stdio.h>
 
@@ -35,8 +36,8 @@
 # endif
 #endif
 
-static struct signals {
-    char *signm;
+static const struct signals {
+    const char *signm;
     int  signo;
 } siglist [] = {
     {"EXIT", 0},
@@ -175,10 +176,9 @@ static struct signals {
 };
 
 static int
-signm2signo(nm)
-    const char *nm;
+signm2signo(const char *nm)
 {
-    struct signals *sigs;
+    const struct signals *sigs;
 
     for (sigs = siglist; sigs->signm; sigs++)
 	if (strcmp(sigs->signm, nm) == 0)
@@ -186,11 +186,10 @@ signm2signo(nm)
     return 0;
 }
 
-static char*
-signo2signm(no)
-    int no;
+static const char*
+signo2signm(int no)
 {
-    struct signals *sigs;
+    const struct signals *sigs;
 
     for (sigs = siglist; sigs->signm; sigs++)
 	if (sigs->signo == no)
@@ -199,8 +198,7 @@ signo2signm(no)
 }
 
 const char *
-ruby_signal_name(no)
-    int no;
+ruby_signal_name(int no)
 {
     return signo2signm(no);
 }
@@ -214,27 +212,24 @@ ruby_signal_name(no)
  */
 
 static VALUE
-esignal_init(argc, argv, self)
-    int argc;
-    VALUE *argv;
-    VALUE self;
+esignal_init(int argc, VALUE *argv, VALUE self)
 {
     int argnum = 1;
     VALUE sig = Qnil;
     int signo;
     const char *signm;
-    char tmpnm[(sizeof(int)*CHAR_BIT)/3+4];
 
     if (argc > 0) {
-	sig = argv[0];
-	if (FIXNUM_P(sig)) argnum = 2;
+	sig = rb_check_to_integer(argv[0], "to_int");
+	if (!NIL_P(sig)) argnum = 2;
+	else sig = argv[0];
     }
     if (argc < 1 || argnum < argc) {
 	rb_raise(rb_eArgError, "wrong number of arguments (%d for %d)",
 		 argc, argnum);
     }
     if (argnum == 2) {
-	signo = FIX2INT(sig);
+	signo = NUM2INT(sig);
 	if (signo < 0 || signo > NSIG) {
 	    rb_raise(rb_eArgError, "invalid signal number (%d)", signo);
 	}
@@ -244,12 +239,11 @@ esignal_init(argc, argv, self)
 	else {
 	    signm = signo2signm(signo);
 	    if (signm) {
-		snprintf(tmpnm, sizeof(tmpnm), "SIG%s", signm);
+		sig = rb_sprintf("SIG%s", signm);
 	    }
 	    else {
-		snprintf(tmpnm, sizeof(tmpnm), "SIG%u", signo);
+		sig = rb_sprintf("SIG%u", signo);
 	    }
-	    sig = rb_str_new2(signm = tmpnm);
 	}
     }
     else {
@@ -259,9 +253,7 @@ esignal_init(argc, argv, self)
 	if (!signo) {
 	    rb_raise(rb_eArgError, "unsupported name `SIG%s'", signm);
 	}
-	if (SYMBOL_P(sig)) {
-	    sig = rb_str_new2(signm);
-	}
+	sig = rb_sprintf("SIG%s", signm);
     }
     rb_call_super(1, &sig);
     rb_iv_set(self, "signo", INT2NUM(signo));
@@ -269,26 +261,35 @@ esignal_init(argc, argv, self)
     return self;
 }
 
-static VALUE
-interrupt_init(self, mesg)
-    VALUE self, mesg;
-{
-    VALUE argv[2];
+/*
+ * call-seq:
+ *    signal_exception.signo   =>  num
+ *
+ *  Returns a signal number.
+ */
 
-    argv[0] = INT2FIX(SIGINT);
-    argv[1] = mesg;
-    return rb_call_super(2, argv);
+static VALUE
+esignal_signo(VALUE self)
+{
+    return rb_iv_get(self, "signo");
+}
+
+static VALUE
+interrupt_init(int argc, VALUE *argv, VALUE self)
+{
+    VALUE args[2];
+
+    args[0] = INT2FIX(SIGINT);
+    rb_scan_args(argc, argv, "01", &args[1]);
+    return rb_call_super(2, args);
 }
 
 void
-ruby_default_signal(sig)
-    int sig;
+ruby_default_signal(int sig)
 {
 #ifndef MACOS_UNUSE_SIGNAL
-    extern rb_pid_t getpid _((void));
-
     signal(sig, SIG_DFL);
-    kill(getpid(), sig);
+    raise(sig);
 #endif
 }
 
@@ -317,14 +318,15 @@ ruby_default_signal(sig)
  */
 
 VALUE
-rb_f_kill(argc, argv)
-    int argc;
-    VALUE *argv;
+rb_f_kill(int argc, VALUE *argv)
 {
+#ifndef HAS_KILLPG
+#define killpg(pg, sig) kill(-(pg), sig)
+#endif
     int negative = 0;
     int sig;
     int i;
-    char *s;
+    const char *s;
 
     rb_secure(2);
     if (argc < 2)
@@ -340,7 +342,7 @@ rb_f_kill(argc, argv)
 	goto str_signal;
 
       case T_STRING:
-	s = RSTRING(argv[0])->ptr;
+	s = RSTRING_PTR(argv[0]);
 	if (s[0] == '-') {
 	    negative++;
 	    s++;
@@ -361,7 +363,7 @@ rb_f_kill(argc, argv)
 
 	    str = rb_check_string_type(argv[0]);
 	    if (!NIL_P(str)) {
-		s = RSTRING(str)->ptr;
+		s = RSTRING_PTR(str);
 		goto str_signal;
 	    }
 	    rb_raise(rb_eArgError, "bad signal type %s",
@@ -373,22 +375,17 @@ rb_f_kill(argc, argv)
     if (sig < 0) {
 	sig = -sig;
 	for (i=1; i<argc; i++) {
-	    int pid = NUM2INT(argv[i]);
-#ifdef HAS_KILLPG
-	    if (killpg(pid, sig) < 0)
-#else
-	    if (kill(-pid, sig) < 0)
-#endif
+	    if (killpg(NUM2PIDT(argv[i]), sig) < 0)
 		rb_sys_fail(0);
 	}
     }
     else {
 	for (i=1; i<argc; i++) {
-	    Check_Type(argv[i], T_FIXNUM);
-	    if (kill(FIX2INT(argv[i]), sig) < 0)
+	    if (kill(NUM2PIDT(argv[i]), sig) < 0)
 		rb_sys_fail(0);
 	}
     }
+    rb_thread_polling();
     return INT2FIX(i-1);
 }
 
@@ -397,13 +394,21 @@ static struct {
     int safe;
 } trap_list[NSIG];
 static rb_atomic_t trap_pending_list[NSIG];
+#if 0
 static char rb_trap_accept_nativethreads[NSIG];
+#endif
 rb_atomic_t rb_trap_pending;
 rb_atomic_t rb_trap_immediate;
 int rb_prohibit_interrupt = 1;
 
+VALUE
+rb_get_trap_cmd(int sig)
+{
+    return trap_list[sig].cmd;
+}
+
 void
-rb_gc_mark_trap_list()
+rb_gc_mark_trap_list(void)
 {
 #ifndef MACOS_UNUSE_SIGNAL
     int i;
@@ -419,67 +424,46 @@ rb_gc_mark_trap_list()
 #define sighandler_t sh_t
 #endif
 
-typedef RETSIGTYPE (*sighandler_t)_((int));
+typedef RETSIGTYPE (*sighandler_t)(int);
 
 #ifdef POSIX_SIGNAL
 static sighandler_t
-ruby_signal(signum, handler)
-    int signum;
-    sighandler_t handler;
+ruby_signal(int signum, sighandler_t handler)
 {
     struct sigaction sigact, old;
 
+#if 0
     rb_trap_accept_nativethreads[signum] = 0;
+#endif
 
-    sigact.sa_handler = handler;
     sigemptyset(&sigact.sa_mask);
+#ifdef SA_SIGINFO
+    sigact.sa_sigaction = (void (*)(int, siginfo_t*, void*))handler;
+    sigact.sa_flags = SA_SIGINFO;
+#else
+    sigact.sa_handler = handler;
     sigact.sa_flags = 0;
-# ifdef SA_NOCLDWAIT
+#endif
+
+#ifdef SA_NOCLDWAIT
     if (signum == SIGCHLD && handler == SIG_IGN)
 	sigact.sa_flags |= SA_NOCLDWAIT;
-# endif
+#endif
     sigaction(signum, &sigact, &old);
     return old.sa_handler;
 }
 
-void
-posix_signal(signum, handler)
-    int signum;
-    sighandler_t handler;
+sighandler_t
+posix_signal(int signum, sighandler_t handler)
 {
-    ruby_signal(signum, handler);
+    return ruby_signal(signum, handler);
 }
 
-# ifdef HAVE_NATIVETHREAD
-static sighandler_t
-ruby_nativethread_signal(signum, handler)
-    int signum;
-    sighandler_t handler;
-{
-    sighandler_t old;
-
-    old = ruby_signal(signum, handler);
-    rb_trap_accept_nativethreads[signum] = 1;
-    return old;
-}
-
-void
-posix_nativethread_signal(signum, handler)
-    int signum;
-    sighandler_t handler;
-{
-    ruby_nativethread_signal(signum, handler);
-}
-# endif
-
 #else /* !POSIX_SIGNAL */
-#define ruby_signal(sig,handler) (rb_trap_accept_nativethreads[sig] = 0, signal((sig),(handler)))
-
-# ifdef HAVE_NATIVETHREAD
+#define ruby_signal(sig,handler) (/* rb_trap_accept_nativethreads[sig] = 0,*/ signal((sig),(handler)))
+#if 0 /* def HAVE_NATIVETHREAD */
 static sighandler_t
-ruby_nativethread_signal(signum, handler)
-    int signum;
-    sighandler_t handler;
+ruby_nativethread_signal(int signum, sighandler_t handler)
 {
     sighandler_t old;
 
@@ -487,18 +471,145 @@ ruby_nativethread_signal(signum, handler)
     rb_trap_accept_nativethreads[signum] = 1;
     return old;
 }
-# endif
-#endif /* POSIX_SIGNAL */
+#endif
+#endif
 
-static void signal_exec _((int sig));
-static void
-signal_exec(sig)
-    int sig;
+static RETSIGTYPE
+sighandler(int sig)
 {
-    if (trap_list[sig].cmd == 0) {
+    rb_vm_t *vm = GET_VM(); /* fix me for Multi-VM */
+    ATOMIC_INC(vm->signal_buff[sig]);
+    ATOMIC_INC(vm->buffered_signal_size);
+
+#if !defined(BSD_SIGNAL) && !defined(POSIX_SIGNAL)
+    ruby_signal(sig, sighandler);
+#endif
+}
+
+#if USE_TRAP_MASK
+# ifdef HAVE_SIGPROCMASK
+static sigset_t trap_last_mask;
+# else
+static int trap_last_mask;
+# endif
+#endif
+
+#if HAVE_PTHREAD_H
+#include <pthread.h>
+#endif
+
+void
+rb_disable_interrupt(void)
+{
+#ifndef _WIN32
+    sigset_t mask;
+    sigfillset(&mask);
+    sigdelset(&mask, SIGVTALRM);
+    sigdelset(&mask, SIGSEGV);
+    pthread_sigmask(SIG_SETMASK, &mask, NULL);
+#endif
+}
+
+void
+rb_enable_interrupt(void)
+{
+#ifndef _WIN32
+    sigset_t mask;
+    sigemptyset(&mask);
+    pthread_sigmask(SIG_SETMASK, &mask, NULL);
+#endif
+}
+
+int
+rb_get_next_signal(rb_vm_t *vm)
+{
+    int i, sig = 0;
+
+    for (i=1; i<RUBY_NSIG; i++) {
+	if (vm->signal_buff[i] > 0) {
+	    rb_disable_interrupt();
+	    {
+		ATOMIC_DEC(vm->signal_buff[i]);
+		ATOMIC_DEC(vm->buffered_signal_size);
+	    }
+	    rb_enable_interrupt();
+	    sig = i;
+	    break;
+	}
+    }
+    return sig;
+}
+
+#ifdef SIGBUS
+static RETSIGTYPE
+sigbus(int sig)
+{
+    rb_bug("Bus Error");
+}
+#endif
+
+#ifdef SIGSEGV
+static int segv_received = 0;
+static RETSIGTYPE
+sigsegv(int sig)
+{
+    if (segv_received) {
+	fprintf(stderr, "SEGV recieved in SEGV handler\n");
+	exit(EXIT_FAILURE);
+    }
+    else {
+	extern int ruby_disable_gc_stress;
+	segv_received = 1;
+	ruby_disable_gc_stress = 1;
+	rb_bug("Segmentation fault");
+    }
+}
+#endif
+
+#ifdef SIGPIPE
+static RETSIGTYPE
+sigpipe(int sig)
+{
+    /* do nothing */
+}
+#endif
+
+static void
+signal_exec(VALUE cmd, int sig)
+{
+    rb_proc_t *proc;
+    VALUE signum = INT2FIX(sig);
+
+    if (TYPE(cmd) == T_STRING) {
+	rb_eval_cmd(cmd, rb_ary_new3(1, signum), trap_list[sig].safe);
+	return;
+    }
+    GetProcPtr(cmd, proc);
+    vm_invoke_proc(GET_THREAD(), proc, proc->block.self, 1, &signum, 0);
+}
+
+void
+rb_trap_exit(void)
+{
+#ifndef MACOS_UNUSE_SIGNAL
+    if (trap_list[0].cmd) {
+	VALUE trap_exit = trap_list[0].cmd;
+
+	trap_list[0].cmd = 0;
+	signal_exec(trap_exit, 0);
+    }
+#endif
+}
+
+void
+rb_signal_exec(rb_thread_t *th, int sig)
+{
+    VALUE cmd = rb_get_trap_cmd(sig);
+
+    if (cmd == 0) {
 	switch (sig) {
 	  case SIGINT:
-	    rb_thread_interrupt();
+	    rb_interrupt();
 	    break;
 #ifdef SIGHUP
 	  case SIGHUP:
@@ -518,142 +629,20 @@ signal_exec(sig)
 #ifdef SIGUSR2
 	  case SIGUSR2:
 #endif
-	    rb_thread_signal_raise(sig);
+	    rb_thread_signal_raise(th, sig);
 	    break;
 	}
     }
-    else if (trap_list[sig].cmd == Qundef) {
-	rb_thread_signal_exit();
+    else if (cmd == Qundef) {
+	rb_thread_signal_exit(th);
     }
     else {
-	rb_thread_trap_eval(trap_list[sig].cmd, sig, trap_list[sig].safe);
+	signal_exec(cmd, sig);
     }
 }
 
-#if defined(HAVE_NATIVETHREAD) && defined(HAVE_NATIVETHREAD_KILL)
-static void
-sigsend_to_ruby_thread(int sig)
-{
-# ifdef HAVE_SIGPROCMASK
-    sigset_t mask, old_mask;
-# else
-    int mask, old_mask;
-# endif
-
-# ifdef HAVE_SIGPROCMASK
-    sigfillset(&mask);
-    sigprocmask(SIG_BLOCK, &mask, &old_mask);
-# else
-    mask = sigblock(~0);
-    sigsetmask(mask);
-# endif
-
-    ruby_native_thread_kill(sig);
-}
-#endif
-
-static RETSIGTYPE sighandler _((int));
-static RETSIGTYPE
-sighandler(sig)
-    int sig;
-{
-#ifdef _WIN32
-#define IN_MAIN_CONTEXT(f, a) (rb_w32_main_context(a, f) ? (void)0 : f(a))
-#else
-#define IN_MAIN_CONTEXT(f, a) f(a)
-#endif
-
-    if (sig >= NSIG) {
-	rb_bug("trap_handler: Bad signal %d", sig);
-    }
-
-#if defined(HAVE_NATIVETHREAD) && defined(HAVE_NATIVETHREAD_KILL)
-    if (!is_ruby_native_thread() && !rb_trap_accept_nativethreads[sig]) {
-        sigsend_to_ruby_thread(sig);
-        return;
-    }
-#endif
-
-#if !defined(BSD_SIGNAL) && !defined(POSIX_SIGNAL)
-    if (rb_trap_accept_nativethreads[sig]) {
-        ruby_nativethread_signal(sig, sighandler);
-    } else {
-        ruby_signal(sig, sighandler);
-    }
-#endif
-
-    if (trap_list[sig].cmd == 0 && ATOMIC_TEST(rb_trap_immediate)) {
-	IN_MAIN_CONTEXT(signal_exec, sig);
-	ATOMIC_SET(rb_trap_immediate, 1);
-    }
-    else {
-	ATOMIC_INC(rb_trap_pending);
-	ATOMIC_INC(trap_pending_list[sig]);
-#ifdef _WIN32
-	rb_w32_interrupted();
-#endif
-    }
-}
-
-#ifdef SIGBUS
-static RETSIGTYPE sigbus _((int));
-static RETSIGTYPE
-sigbus(sig)
-    int sig;
-{
-#if defined(HAVE_NATIVETHREAD) && defined(HAVE_NATIVETHREAD_KILL)
-    if (!is_ruby_native_thread() && !rb_trap_accept_nativethreads[sig]) {
-        sigsend_to_ruby_thread(sig);
-        return;
-    }
-#endif
-
-    rb_bug("Bus Error");
-}
-#endif
-
-#ifdef SIGSEGV
-static RETSIGTYPE sigsegv _((int));
-static RETSIGTYPE
-sigsegv(sig)
-    int sig;
-{
-#if defined(HAVE_NATIVETHREAD) && defined(HAVE_NATIVETHREAD_KILL)
-    if (!is_ruby_native_thread() && !rb_trap_accept_nativethreads[sig]) {
-        sigsend_to_ruby_thread(sig);
-        return;
-    }
-#endif
-
-    rb_bug("Segmentation fault");
-}
-#endif
-
-#ifdef SIGPIPE
-static RETSIGTYPE sigpipe _((int));
-static RETSIGTYPE
-sigpipe(sig)
-    int sig;
-{
-    /* do nothing */
-}
-#endif
-
-void
-rb_trap_exit()
-{
-#ifndef MACOS_UNUSE_SIGNAL
-    if (trap_list[0].cmd) {
-	VALUE trap_exit = trap_list[0].cmd;
-
-	trap_list[0].cmd = 0;
-	rb_eval_cmd(trap_exit, rb_ary_new3(1, INT2FIX(0)), trap_list[0].safe);
-    }
-#endif
-}
-
 void
-rb_trap_exec()
+rb_trap_exec(void)
 {
 #ifndef MACOS_UNUSE_SIGNAL
     int i;
@@ -661,7 +650,7 @@ rb_trap_exec()
     for (i=0; i<NSIG; i++) {
 	if (trap_pending_list[i]) {
 	    trap_pending_list[i] = 0;
-	    signal_exec(i);
+	    rb_signal_exec(GET_THREAD(), i);
 	}
     }
 #endif /* MACOS_UNUSE_SIGNAL */
@@ -676,80 +665,141 @@ struct trap_arg {
     int mask;
 # endif
 #endif
-    VALUE sig, cmd;
+    int sig;
+    sighandler_t func;
+    VALUE cmd;
 };
 
-# ifdef HAVE_SIGPROCMASK
-static sigset_t trap_last_mask;
-# else
-static int trap_last_mask;
-# endif
-
-static RETSIGTYPE sigexit _((int));
-static RETSIGTYPE
-sigexit(sig)
-    int sig;
+static sighandler_t
+default_handler(int sig)
 {
-    rb_thread_signal_exit();
+    sighandler_t func;
+    switch (sig) {
+      case SIGINT:
+#ifdef SIGHUP
+      case SIGHUP:
+#endif
+#ifdef SIGQUIT
+      case SIGQUIT:
+#endif
+#ifdef SIGTERM
+      case SIGTERM:
+#endif
+#ifdef SIGALRM
+      case SIGALRM:
+#endif
+#ifdef SIGUSR1
+      case SIGUSR1:
+#endif
+#ifdef SIGUSR2
+      case SIGUSR2:
+#endif
+        func = sighandler;
+        break;
+#ifdef SIGBUS
+      case SIGBUS:
+        func = sigbus;
+        break;
+#endif
+#ifdef SIGSEGV
+      case SIGSEGV:
+        func = sigsegv;
+        break;
+#endif
+#ifdef SIGPIPE
+      case SIGPIPE:
+        func = sigpipe;
+        break;
+#endif
+      default:
+        func = SIG_DFL;
+        break;
+    }
+
+    return func;
 }
 
-static VALUE
-trap(arg)
-    struct trap_arg *arg;
+static sighandler_t
+trap_handler(VALUE *cmd, int sig)
 {
-    sighandler_t func, oldfunc;
-    VALUE command, oldcmd;
-    int sig = -1;
-    char *s;
+    sighandler_t func = sighandler;
+    VALUE command;
 
-    func = sighandler;
-    command = arg->cmd;
-    if (NIL_P(command)) {
+    if (NIL_P(*cmd)) {
 	func = SIG_IGN;
     }
-    else if (TYPE(command) == T_STRING) {
-	SafeStringValue(command);	/* taint check */
-	if (RSTRING(command)->len == 0) {
-	    func = SIG_IGN;
-	}
-	else if (RSTRING(command)->len == 7) {
-	    if (strncmp(RSTRING(command)->ptr, "SIG_IGN", 7) == 0) {
-		func = SIG_IGN;
-	    }
-	    else if (strncmp(RSTRING(command)->ptr, "SIG_DFL", 7) == 0) {
-		func = SIG_DFL;
-	    }
-	    else if (strncmp(RSTRING(command)->ptr, "DEFAULT", 7) == 0) {
-		func = SIG_DFL;
+    else {
+	command = rb_check_string_type(*cmd);
+	if (!NIL_P(command)) {
+	    SafeStringValue(command);	/* taint check */
+	    *cmd = command;
+	    switch (RSTRING_LEN(command)) {
+	      case 0:
+                goto sig_ign;
+		break;
+              case 14:
+		if (strncmp(RSTRING_PTR(command), "SYSTEM_DEFAULT", 14) == 0) {
+                    func = SIG_DFL;
+                    *cmd = 0;
+		}
+                break;
+	      case 7:
+		if (strncmp(RSTRING_PTR(command), "SIG_IGN", 7) == 0) {
+sig_ign:
+                    func = SIG_IGN;
+                    *cmd = 0;
+		}
+		else if (strncmp(RSTRING_PTR(command), "SIG_DFL", 7) == 0) {
+sig_dfl:
+                    func = default_handler(sig);
+                    *cmd = 0;
+		}
+		else if (strncmp(RSTRING_PTR(command), "DEFAULT", 7) == 0) {
+                    goto sig_dfl;
+		}
+		break;
+	      case 6:
+		if (strncmp(RSTRING_PTR(command), "IGNORE", 6) == 0) {
+                    goto sig_ign;
+		}
+		break;
+	      case 4:
+		if (strncmp(RSTRING_PTR(command), "EXIT", 4) == 0) {
+		    *cmd = Qundef;
+		}
+		break;
 	    }
 	}
-	else if (RSTRING(command)->len == 6) {
-	    if (strncmp(RSTRING(command)->ptr, "IGNORE", 6) == 0) {
-		func = SIG_IGN;
-	    }
+	else {
+	    rb_proc_t *proc;
+	    GetProcPtr(*cmd, proc);
 	}
-	else if (RSTRING(command)->len == 4) {
-	    if (strncmp(RSTRING(command)->ptr, "EXIT", 4) == 0) {
-		func = sigexit;
-	    }
-	}
-    }
-    if (func == SIG_IGN || func == SIG_DFL) {
-	command = 0;
     }
 
-    switch (TYPE(arg->sig)) {
+    return func;
+}
+
+static int
+trap_signm(VALUE vsig)
+{
+    int sig = -1;
+    const char *s;
+
+    switch (TYPE(vsig)) {
       case T_FIXNUM:
-	sig = FIX2INT(arg->sig);
+	sig = FIX2INT(vsig);
+	if (sig < 0 || sig >= NSIG) {
+	    rb_raise(rb_eArgError, "invalid signal number (%d)", sig);
+	}
 	break;
 
       case T_SYMBOL:
-	s = rb_id2name(SYM2ID(arg->sig));
+	s = rb_id2name(SYM2ID(vsig));
 	if (!s) rb_raise(rb_eArgError, "bad signal");
 	goto str_signal;
 
-      case T_STRING:
-	s = RSTRING(arg->sig)->ptr;
+      default:
+	s = StringValuePtr(vsig);
 
       str_signal:
 	if (strncmp("SIG", s, 3) == 0)
@@ -758,65 +808,31 @@ trap(arg)
 	if (sig == 0 && strcmp(s, "EXIT") != 0)
 	    rb_raise(rb_eArgError, "unsupported signal SIG%s", s);
     }
+    return sig;
+}
+
+static VALUE
+trap(struct trap_arg *arg)
+{
+    sighandler_t oldfunc, func = arg->func;
+    VALUE oldcmd, command = arg->cmd;
+    int sig = arg->sig;
 
-    if (sig < 0 || sig >= NSIG) {
-	rb_raise(rb_eArgError, "invalid signal number (%d)", sig);
-    }
-#if defined(HAVE_SETITIMER)
-    if (sig == SIGVTALRM) {
-	rb_raise(rb_eArgError, "SIGVTALRM reserved for Thread; can't set handler");
-    }
-#endif
-    if (func == SIG_DFL) {
-	switch (sig) {
-	  case SIGINT:
-#ifdef SIGHUP
-	  case SIGHUP:
-#endif
-#ifdef SIGQUIT
-	  case SIGQUIT:
-#endif
-#ifdef SIGTERM
-	  case SIGTERM:
-#endif
-#ifdef SIGALRM
-	  case SIGALRM:
-#endif
-#ifdef SIGUSR1
-	  case SIGUSR1:
-#endif
-#ifdef SIGUSR2
-	  case SIGUSR2:
-#endif
-	    func = sighandler;
-	    break;
-#ifdef SIGBUS
-	  case SIGBUS:
-	    func = sigbus;
-	    break;
-#endif
-#ifdef SIGSEGV
-	  case SIGSEGV:
-	    func = sigsegv;
-	    break;
-#endif
-#ifdef SIGPIPE
-	  case SIGPIPE:
-	    func = sigpipe;
-	    break;
-#endif
-	}
-    }
     oldfunc = ruby_signal(sig, func);
     oldcmd = trap_list[sig].cmd;
-    if (!oldcmd) {
+    switch (oldcmd) {
+      case 0:
 	if (oldfunc == SIG_IGN) oldcmd = rb_str_new2("IGNORE");
 	else if (oldfunc == sighandler) oldcmd = rb_str_new2("DEFAULT");
 	else oldcmd = Qnil;
+	break;
+      case Qundef:
+	oldcmd = rb_str_new2("EXIT");
+	break;
     }
 
     trap_list[sig].cmd = command;
-    trap_list[sig].safe = ruby_safe_level;
+    trap_list[sig].safe = rb_safe_level();
     /* enable at least specified signal. */
 #if USE_TRAP_MASK
 #ifdef HAVE_SIGPROCMASK
@@ -830,8 +846,7 @@ trap(arg)
 
 #if USE_TRAP_MASK
 static VALUE
-trap_ensure(arg)
-    struct trap_arg *arg;
+trap_ensure(struct trap_arg *arg)
 {
     /* enable interrupt */
 #ifdef HAVE_SIGPROCMASK
@@ -845,7 +860,7 @@ trap_ensure(arg)
 #endif
 
 void
-rb_trap_restore_mask()
+rb_trap_restore_mask(void)
 {
 #if USE_TRAP_MASK
 # ifdef HAVE_SIGPROCMASK
@@ -858,19 +873,22 @@ rb_trap_restore_mask()
 
 /*
  * call-seq:
- *   Signal.trap( signal, proc ) => obj
+ *   Signal.trap( signal, command ) => obj
  *   Signal.trap( signal ) {| | block } => obj
  *
  * Specifies the handling of signals. The first parameter is a signal
  * name (a string such as ``SIGALRM'', ``SIGUSR1'', and so on) or a
  * signal number. The characters ``SIG'' may be omitted from the
  * signal name. The command or block specifies code to be run when the
- * signal is raised. If the command is the string ``IGNORE'' or
- * ``SIG_IGN'', the signal will be ignored. If the command is
- * ``DEFAULT'' or ``SIG_DFL'', the operating system's default handler
- * will be invoked. If the command is ``EXIT'', the script will be
- * terminated by the signal. Otherwise, the given command or block
- * will be run.
+ * signal is raised.
+ * If the command is the string ``IGNORE'' or ``SIG_IGN'', the signal
+ * will be ignored.
+ * If the command is ``DEFAULT'' or ``SIG_DFL'', the Ruby's default handler
+ * will be invoked.
+ * If the command is ``EXIT'', the script will be terminated by the signal.
+ * If the command is ``SYSTEM_DEFAULT'', the operating system's default
+ * handler will be invoked.
+ * Otherwise, the given command or block will be run.
  * The special signal name ``EXIT'' or signal number zero will be
  * invoked just prior to program termination.
  * trap returns the previous handler for the given signal.
@@ -885,9 +903,7 @@ rb_trap_restore_mask()
  *     Terminating: 27460
  */
 static VALUE
-sig_trap(argc, argv)
-    int argc;
-    VALUE *argv;
+sig_trap(int argc, VALUE *argv)
 {
     struct trap_arg arg;
 
@@ -896,12 +912,14 @@ sig_trap(argc, argv)
 	rb_raise(rb_eArgError, "wrong number of arguments -- trap(sig, cmd)/trap(sig){...}");
     }
 
-    arg.sig = argv[0];
+    arg.sig = trap_signm(argv[0]);
     if (argc == 1) {
 	arg.cmd = rb_block_proc();
+	arg.func = sighandler;
     }
     else if (argc == 2) {
 	arg.cmd = argv[1];
+	arg.func = trap_handler(&arg.cmd, arg.sig);
     }
 
     if (OBJ_TAINTED(arg.cmd)) {
@@ -929,13 +947,13 @@ sig_trap(argc, argv)
  * Returns a list of signal names mapped to the corresponding
  * underlying signal numbers.
  *
- * Signal.list   #=> {"ABRT"=>6, "ALRM"=>14, "BUS"=>7, "CHLD"=>17, "CLD"=>17, "CONT"=>18, "FPE"=>8, "HUP"=>1, "ILL"=>4, "INT"=>2, "IO"=>29, "IOT"=>6, "KILL"=>9, "PIPE"=>13, "POLL"=>29, "PROF"=>27, "PWR"=>30, "QUIT"=>3, "SEGV"=>11, "STOP"=>19, "SYS"=>31, "TERM"=>15, "TRAP"=>5, "TSTP"=>20, "TTIN"=>21, "TTOU"=>22, "URG"=>23, "USR1"=>10, "USR2"=>12, "VTALRM"=>26, "WINCH"=>28, "XCPU"=>24, "XFSZ"=>25}
+ *   Signal.list   #=> {"EXIT"=>0, "HUP"=>1, "INT"=>2, "QUIT"=>3, "ILL"=>4, "TRAP"=>5, "IOT"=>6, "ABRT"=>6, "FPE"=>8, "KILL"=>9, "BUS"=>7, "SEGV"=>11, "SYS"=>31, "PIPE"=>13, "ALRM"=>14, "TERM"=>15, "URG"=>23, "STOP"=>19, "TSTP"=>20, "CONT"=>18, "CHLD"=>17, "CLD"=>17, "TTIN"=>21, "TTOU"=>22, "IO"=>29, "XCPU"=>24, "XFSZ"=>25, "VTALRM"=>26, "PROF"=>27, "WINCH"=>28, "USR1"=>10, "USR2"=>12, "PWR"=>30, "POLL"=>29}
  */
 static VALUE
-sig_list()
+sig_list(void)
 {
     VALUE h = rb_hash_new();
-    struct signals *sigs;
+    const struct signals *sigs;
 
     for (sigs = siglist; sigs->signm; sigs++) {
 	rb_hash_aset(h, rb_str_new2(sigs->signm), INT2FIX(sigs->signo));
@@ -944,9 +962,7 @@ sig_list()
 }
 
 static void
-install_sighandler(signum, handler)
-    int signum;
-    sighandler_t handler;
+install_sighandler(int signum, sighandler_t handler)
 {
     sighandler_t old;
 
@@ -956,38 +972,9 @@ install_sighandler(signum, handler)
     }
 }
 
-#if 0
-/* 
- *   If you write a handler which works on any native thread
- *   (even if the thread is NOT a ruby's one), please enable
- *   this function and use it to install the handler, instead
- *   of `install_sighandler()'.
- */
-#ifdef HAVE_NATIVETHREAD
+#if defined(SIGCLD) || defined(SIGCHLD)
 static void
-install_nativethread_sighandler(signum, handler)
-    int signum;
-    sighandler_t handler;
-{
-    sighandler_t old;
-    int old_st;
-
-    old_st = rb_trap_accept_nativethreads[signum];
-    old = ruby_nativethread_signal(signum, handler);
-    if (old != SIG_DFL) {
-        if (old_st) {
-            ruby_nativethread_signal(signum, old);
-        } else {
-            ruby_signal(signum, old);
-        }
-    }
-}
-#endif
-#endif
-
-static void
-init_sigchld(sig)
-    int sig;
+init_sigchld(int sig)
 {
     sighandler_t oldfunc;
 #if USE_TRAP_MASK
@@ -1026,6 +1013,23 @@ init_sigchld(sig)
     trap_last_mask = mask;
 #endif
 }
+#endif
+
+void
+ruby_sig_finalize()
+{
+    sighandler_t oldfunc;
+
+    oldfunc = ruby_signal(SIGINT, SIG_IGN);
+    if (oldfunc == sighandler) {
+	ruby_signal(SIGINT, SIG_DFL);
+    }
+}
+
+
+#ifdef RUBY_DEBUG_ENV
+int ruby_enable_coredump = 0;
+#endif
 
 /*
  * Many operating systems allow signals to be sent to running
@@ -1065,7 +1069,7 @@ init_sigchld(sig)
  * systems; in particular signal delivery may not always be reliable.
  */
 void
-Init_signal()
+Init_signal(void)
 {
 #ifndef MACOS_UNUSE_SIGNAL
     VALUE mSignal = rb_define_module("Signal");
@@ -1075,9 +1079,9 @@ Init_signal()
     rb_define_module_function(mSignal, "list", sig_list, 0);
 
     rb_define_method(rb_eSignal, "initialize", esignal_init, -1);
-    rb_attr(rb_eSignal, rb_intern("signo"), 1, 0, 0);
+    rb_define_method(rb_eSignal, "signo", esignal_signo, 0);
     rb_alias(rb_eSignal, rb_intern("signm"), rb_intern("message"));
-    rb_define_method(rb_eInterrupt, "initialize", interrupt_init, 1);
+    rb_define_method(rb_eInterrupt, "initialize", interrupt_init, -1);
 
     install_sighandler(SIGINT, sighandler);
 #ifdef SIGHUP
@@ -1099,12 +1103,17 @@ Init_signal()
     install_sighandler(SIGUSR2, sighandler);
 #endif
 
+#ifdef RUBY_DEBUG_ENV
+    if (!ruby_enable_coredump)
+#endif
+    {
 #ifdef SIGBUS
     install_sighandler(SIGBUS, sigbus);
 #endif
 #ifdef SIGSEGV
     install_sighandler(SIGSEGV, sigsegv);
 #endif
+    }
 #ifdef SIGPIPE
     install_sighandler(SIGPIPE, sigpipe);
 #endif
