@@ -493,6 +493,33 @@ class TestM17N < Test::Unit::TestCase
     assert_match(/[[:space:]]/, "\u{00a0}")
   end
 
+  def test_regexp_property
+    s = '\p{Hiragana}'.force_encoding("euc-jp")
+    assert_equal(Encoding::EUC_JP, s.encoding)
+    r = nil
+    assert_nothing_raised {
+      r = Regexp.new(s)
+    }
+    assert(r.fixed_encoding?)
+    assert_match(r, "\xa4\xa2".force_encoding("euc-jp"))
+
+    r = eval('/\p{Hiragana}/'.force_encoding("euc-jp"))
+    assert(r.fixed_encoding?)
+    assert_match(r, "\xa4\xa2".force_encoding("euc-jp"))
+
+    r = /\p{Hiragana}/e
+    assert(r.fixed_encoding?)
+    assert_match(r, "\xa4\xa2".force_encoding("euc-jp"))
+
+    r = eval('/\u{3042}\p{Hiragana}/'.force_encoding("euc-jp"))
+    assert(r.fixed_encoding?)
+    assert_equal(Encoding::UTF_8, r.encoding)
+
+    r = eval('/\p{Hiragana}\u{3042}/'.force_encoding("euc-jp"))
+    assert(r.fixed_encoding?)
+    assert_equal(Encoding::UTF_8, r.encoding)
+  end
+
   def test_regexp_embed_preprocess
     r1 = /\xa4\xa2/e
     r2 = /#{r1}/
@@ -810,6 +837,19 @@ class TestM17N < Test::Unit::TestCase
     assert_equal(false, str[0..-1].ascii_only?)
   end
 
+  def test_utf8str_aref
+    s = "abcdefghijklmnopqrstuvwxyz\u{3042 3044 3046 3048 304A}"
+    assert_equal("a", s[0])
+    assert_equal("h", s[7])
+    assert_equal("i", s[8])
+    assert_equal("j", s[9])
+    assert_equal("\u{3044}", s[27])
+    assert_equal("\u{3046}", s[28])
+    assert_equal("\u{3048}", s[29])
+    s = "abcdefghijklmnopqrstuvw\u{3042 3044 3046 3048 304A}"
+    assert_equal("\u{3044}", s[24])
+  end
+
   def test_str_aref_len
     assert_equal(a("\xa1"), a("\xc2\xa1\xc2\xa2\xc2\xa3")[1, 1])
     assert_equal(a("\xa1\xc2"), a("\xc2\xa1\xc2\xa2\xc2\xa3")[1, 2])
@@ -864,6 +904,9 @@ class TestM17N < Test::Unit::TestCase
     }
 
     assert_equal(e("\xA1\xA1"), a("a").tr(a("a"), e("\xA1\xA1")))
+
+    assert_equal("X\u3042\u3044X", "A\u3042\u3044\u3046".tr("^\u3042\u3044", "X"))
+    assert_equal("\u3042\u3046" * 100, ("\u3042\u3044" * 100).tr("\u3044", "\u3046"))
   end
 
   def test_tr_s
@@ -881,6 +924,10 @@ class TestM17N < Test::Unit::TestCase
     assert_equal(1, e("\xa1\xa2").delete("z").length)
     s = e("\xa3\xb0\xa3\xb1\xa3\xb2\xa3\xb3\xa3\xb4")
     assert_raise(ArgumentError){s.delete(a("\xa3\xb2"))}
+
+    a = "\u3042\u3044\u3046\u3042\u3044\u3046"
+    a.delete!("\u3042\u3044", "^\u3044")
+    assert_equal("\u3044\u3046\u3044\u3046", a)
   end
 
   def test_include?
@@ -951,6 +998,12 @@ class TestM17N < Test::Unit::TestCase
     assert_equal(u("\xf0jihgfedcba"), u("abcdefghij\xf0").reverse)
   end
 
+  def test_reverse_bang
+    s = u("abcdefghij\xf0")
+    s.reverse!
+    assert_equal(u("\xf0jihgfedcba"), s)
+  end
+
   def test_plus
     assert_raise(ArgumentError){u("\xe3\x81\x82") + a("\xa1")}
   end
@@ -1028,8 +1081,8 @@ class TestM17N < Test::Unit::TestCase
      mu = method(u("\xc2\xa1"))
      assert_not_equal(me.name, mu.name)
      assert_not_equal(me.inspect, mu.inspect)
-     assert_equal(e("\xc2\xa1"), me.name)
-     assert_equal(u("\xc2\xa1"), mu.name)
+     assert_equal(e("\xc2\xa1"), me.name.to_s)
+     assert_equal(u("\xc2\xa1"), mu.name.to_s)
   end
 
   def test_symbol
@@ -1184,8 +1237,36 @@ class TestM17N < Test::Unit::TestCase
     assert_equal(true, (s.dup << s).valid_encoding?)
     assert_equal(true, "".center(2, s).valid_encoding?)
 
-     s = "\xa1\xa1\x8f".force_encoding("euc-jp")
-     assert_equal(false, s.valid_encoding?)
-     assert_equal(true, s.reverse.valid_encoding?)
-   end
+    s = "\xa1\xa1\x8f".force_encoding("euc-jp")
+    assert_equal(false, s.valid_encoding?)
+    assert_equal(true, s.reverse.valid_encoding?)
+  end
+
+  def test_getbyte
+    assert_equal(0x82, u("\xE3\x81\x82\xE3\x81\x84").getbyte(2))
+    assert_equal(0x82, u("\xE3\x81\x82\xE3\x81\x84").getbyte(-4))
+    assert_nil(u("\xE3\x81\x82\xE3\x81\x84").getbyte(100))
+  end
+
+  def test_setbyte
+    s = u("\xE3\x81\x82\xE3\x81\x84")
+    s.setbyte(2, 0x84)
+    assert_equal(u("\xE3\x81\x84\xE3\x81\x84"), s)
+
+    s = u("\xE3\x81\x82\xE3\x81\x84")
+    assert_raise(IndexError) { s.setbyte(100, 0) }
+
+    s = u("\xE3\x81\x82\xE3\x81\x84")
+    s.setbyte(-4, 0x84)
+    assert_equal(u("\xE3\x81\x84\xE3\x81\x84"), s)
+  end
+
+  def test_compatible
+    assert_equal(nil, Encoding.compatible?("",0), "moved from btest/knownbug")
+  end
+
+  def test_force_encoding
+    assert(("".center(1, "\x80".force_encoding("utf-8")); true),
+           "moved from btest/knownbug, [ruby-dev:33807]")
+  end
 end
